#include "stdio.h"
#include "const.h"
#include "math.h"
#include "string.h"
#include "stdlib.h"
#include "float.h"
#include "time.h"
//#include "omp.h"
#include "unistd.h"
#include "sys/types.h"

float sigmoid(float x)
{
	return (1.0F-expf(-1*x))/(1.0F+expf(-1*x));
}


void neuralnetwork(float output[],float weight[],float sample[])
{
	float hidden[HIDDEN_LAYER_NUMBER][HIDDEN_NEURONS_NUMBER];
	for(int i=0;i<HIDDEN_NEURONS_NUMBER;i++)
	{
		float sum=0;
		for(int j=0;j<INPUT_NEURONS_NUMBER;j++)
			sum+=weight[i*(INPUT_NEURONS_NUMBER+1)+j]*sample[j];
		hidden[0][i]=sigmoid(sum+weight[i*(INPUT_NEURONS_NUMBER+1)+INPUT_NEURONS_NUMBER]);
	}

	int startpoint=HIDDEN_NEURONS_NUMBER*(INPUT_NEURONS_NUMBER+1);
	for(int l=1;l<HIDDEN_LAYER_NUMBER;l++)
	{
		for(int i=0;i<HIDDEN_NEURONS_NUMBER;i++)
		{
			float sum=0;
			for(int j=0;j<HIDDEN_NEURONS_NUMBER;j++)
				sum+=weight[startpoint+i*(HIDDEN_NEURONS_NUMBER+1)+j]*hidden[l-1][j];
			hidden[l][i]=sigmoid(sum+weight[startpoint+i*(HIDDEN_NEURONS_NUMBER+1)+HIDDEN_NEURONS_NUMBER]);

		}
		startpoint+=(HIDDEN_NEURONS_NUMBER+1)*HIDDEN_NEURONS_NUMBER;
	}

	for(int i=0;i<OUTPUT_NEURONS_NUMBER;i++)
	{
		float sum=0;
		for(int j=0;j<HIDDEN_NEURONS_NUMBER;j++)
			sum+=weight[startpoint+i*(HIDDEN_NEURONS_NUMBER+1)+j]*hidden[HIDDEN_LAYER_NUMBER-1][j];
		output[i]=sigmoid(sum+weight[startpoint+i*(HIDDEN_NEURONS_NUMBER+1)+HIDDEN_NEURONS_NUMBER]);
	}

}

float distance(float x[], float y[],int dimension)
{
	float sum=0;
	for(int i=0;i<dimension;i++)
		sum+=(x[i]-y[i])*(x[i]-y[i]);
	return sqrtf(sum);
}



int ReadFile(char filename[],float data[][INPUT_DIMENSION+1])
{
	char s[64000];
	char *p; 
	int dimension=0;

	FILE *fp=fopen(filename,"r");
	int d,number=0;
	fseek(fp,0,SEEK_SET);

	//printf("start read %s\n",filename);
	fscanf(fp,"%s",s);
	while(strcmp(s,"end")!=0)
	{	
		d=0;
		p=strtok(s,",|	| |;");
		while(p)
		{
			data[number][d++]=atof(p);
			p=strtok(NULL,",|	| |;");
		}
		fscanf(fp,"%s",s);
		number++;
	}
	fclose(fp);
	//printf("stop read %s\n",filename);

	return number;
}

int GetDataWeight(float data[][INPUT_DIMENSION+1],int num,float weight[])
{
	int count[CLASS_NUMBER];
	for(int i=0;i<CLASS_NUMBER;i++)
		count[i]=0;

	for(int i=0;i<num;i++)
	{
		int classid=(int)(data[i][INPUT_DIMENSION]+0.1);
		count[classid]++;
	}

	for(int i=0;i<CLASS_NUMBER;i++)
		weight[i]=((float)num/CLASS_NUMBER)/count[i];
}


void normalize(float colorpoints[][PARTITION_SPACE_DIMENSION+1],int datanum,float weight[],float outave[],float outstd[])
{
	for(int i=0;i<PARTITION_SPACE_DIMENSION;i++)
	{
		outave[i]=0;
		outstd[i]=0;
	}


	for(int i=0;i<PARTITION_SPACE_DIMENSION;i++)
	{
		float sum=0;
		for(int j=0;j<datanum;j++)
		{
			float w=weight[(int)(colorpoints[j][PARTITION_SPACE_DIMENSION]+0.1)];
			outave[i]+=w*colorpoints[j][i];
			sum+=w;
		}
		outave[i]=outave[i]/sum;
	}

		
	for(int i=0;i<PARTITION_SPACE_DIMENSION;i++)
	{
		float sum=0;
		for(int j=0;j<datanum;j++)
		{		
			float w=weight[(int)(colorpoints[j][PARTITION_SPACE_DIMENSION]+0.1)];
			outstd[i]+=w*(colorpoints[j][i]-outave[i])*(colorpoints[j][i]-outave[i]);
			sum+=w;
		}
		outstd[i]=sqrtf(outstd[i]/sum);
	}
}



void map(float x[],float data[][INPUT_DIMENSION+1],int datanum,float colorpoints[][PARTITION_SPACE_DIMENSION+1])
{
	for(int i=0;i<datanum;i++)
	{
		neuralnetwork(colorpoints[i],x,data[i]);
		colorpoints[i][PARTITION_SPACE_DIMENSION]=data[i][INPUT_DIMENSION];
	}
}



float fitnessf(float x[],float weight[],float data[][INPUT_DIMENSION+1],int datanum)
{
	float minweight=FLT_MAX;
	for(int i=0;i<CLASS_NUMBER;i++)
		if(weight[i]<minweight)
			minweight=weight[i];


	float	(*colorpoints)[PARTITION_SPACE_DIMENSION+1]=
		new float[datanum][PARTITION_SPACE_DIMENSION+1];

	map(x,data,datanum,colorpoints);

	float ave[PARTITION_SPACE_DIMENSION];
	float std[PARTITION_SPACE_DIMENSION];
	normalize(colorpoints,datanum,weight,ave,std);
	for(int i=0;i<PARTITION_SPACE_DIMENSION;i++)
		for(int j=0;j<datanum;j++)
			colorpoints[j][i]=(colorpoints[j][i]-ave[i])/std[i];

	for(int j=0;j<datanum;j++)
	{
		float d=0;
		for(int i=0;i<PARTITION_SPACE_DIMENSION;i++)
			d+=colorpoints[j][i]*colorpoints[j][i];
		d=sqrtf(d);

		float d2=(expf(d/BOUND_WEIGHT)-expf(-1*d/BOUND_WEIGHT))/(expf(d/BOUND_WEIGHT)+expf(-1*d/BOUND_WEIGHT));
		float ratio=d2/d;
		for(int i=0;i<PARTITION_SPACE_DIMENSION;i++)	
			colorpoints[j][i]=colorpoints[j][i]*ratio;
		
	}
	


	float **SimilarityArray=new float*[datanum];
	for(long i=0;i<datanum;i++)
		SimilarityArray[i]= new float[datanum];

	for(int i=0;i<datanum;i++)
		for(int j=0;j<datanum;j++)
		{
			if(i==j)
				SimilarityArray[i][j]=2.0F;
			else
				SimilarityArray[i][j]=distance(colorpoints[i],colorpoints[j],PARTITION_SPACE_DIMENSION);
		}


	float count=0;

	for(int i=0;i<datanum;i++)
	{

		int *index=new int[datanum];
		for(int j=0;j<datanum;j++)
			index[j]=j;

		for(int j=0;j<(int)(SPARSE_NEIGHBOUR_NUMBER/minweight)+2;j++)//should be carefully tuned
			for(int k=datanum-1;k>j;k--)
			{
				if(SimilarityArray[i][index[k]]<SimilarityArray[i][index[k-1]])
				{
					int temp=index[k];
					index[k]=index[k-1];
					index[k-1]=temp;
				}
			}

		float sum=0;
		int j;
		float dminself=0;
		float dminnonself=0;

		for(j=0;sum+weight[(int)(colorpoints[index[j]][PARTITION_SPACE_DIMENSION]+0.1F)]<SPARSE_NEIGHBOUR_NUMBER;j++)
		{

			float w=weight[(int)(colorpoints[index[j]][PARTITION_SPACE_DIMENSION]+0.1)];

			if((int)(colorpoints[i][PARTITION_SPACE_DIMENSION]+0.1F)!=(int)(colorpoints[index[j]][PARTITION_SPACE_DIMENSION]+0.1F))
				dminnonself+=w*(2.0F-SimilarityArray[i][index[j]]);
			else
				dminself+=w*(2.0F-SimilarityArray[i][index[j]]);
			
			sum+=w;
		}

		if((int)(colorpoints[i][PARTITION_SPACE_DIMENSION]+0.1F)!=(int)(colorpoints[index[j]][PARTITION_SPACE_DIMENSION]+0.1F))
			dminnonself+=(SPARSE_NEIGHBOUR_NUMBER-sum)*(2.0F-SimilarityArray[i][index[j]]);
		else
			dminself+=(SPARSE_NEIGHBOUR_NUMBER-sum)*(2.0F-SimilarityArray[i][index[j]]);

		//count+=dminnonself-A*dminself;
		count+=weight[(int)(colorpoints[i][PARTITION_SPACE_DIMENSION]+0.1)]*(dminnonself-A*dminself);
	
		delete []index;
	}


	for(long i=0;i<datanum;i++)
		delete []SimilarityArray[i];
	delete []SimilarityArray;

	delete []colorpoints;

	return count;

}

void test(float x[],float weight[],float traindata[][INPUT_DIMENSION+1],int traindatanum,float testdata[][INPUT_DIMENSION+1],int testdatanum,float measurement[],int neigh_num)
{
	float minweight=FLT_MAX;
	for(int i=0;i<CLASS_NUMBER;i++)
		if(weight[i]<minweight)
			minweight=weight[i];

	float	(*traincolorpoints)[PARTITION_SPACE_DIMENSION+1]=
		new float[traindatanum][PARTITION_SPACE_DIMENSION+1];
	map(x,traindata,traindatanum,traincolorpoints);
	
	float ave[PARTITION_SPACE_DIMENSION];
	float std[PARTITION_SPACE_DIMENSION];
	normalize(traincolorpoints,traindatanum,weight,ave,std);
	for(int i=0;i<PARTITION_SPACE_DIMENSION;i++)
		for(int j=0;j<traindatanum;j++)
			traincolorpoints[j][i]=(traincolorpoints[j][i]-ave[i])/std[i];

	for(int j=0;j<traindatanum;j++)
	{
		float d=0;
		for(int i=0;i<PARTITION_SPACE_DIMENSION;i++)
			d+=traincolorpoints[j][i]*traincolorpoints[j][i];
		d=sqrtf(d);

		float d2=(expf(d/BOUND_WEIGHT)-expf(-1*d/BOUND_WEIGHT))/(expf(d/BOUND_WEIGHT)+expf(-1*d/BOUND_WEIGHT));
		float ratio=d2/d;
		for(int i=0;i<PARTITION_SPACE_DIMENSION;i++)	
			traincolorpoints[j][i]=traincolorpoints[j][i]*ratio;

	}



	float	(*testcolorpoints)[PARTITION_SPACE_DIMENSION+1]=
		new float[testdatanum][PARTITION_SPACE_DIMENSION+1];
	map(x,testdata,testdatanum,testcolorpoints);
	for(int i=0;i<PARTITION_SPACE_DIMENSION;i++)
		for(int j=0;j<testdatanum;j++)
			testcolorpoints[j][i]=(testcolorpoints[j][i]-ave[i])/std[i];

	for(int j=0;j<testdatanum;j++)
	{
		float d=0;
		for(int i=0;i<PARTITION_SPACE_DIMENSION;i++)
			d+=testcolorpoints[j][i]*testcolorpoints[j][i];
		d=sqrtf(d);

		float d2=(expf(d/BOUND_WEIGHT)-expf(-1*d/BOUND_WEIGHT))/(expf(d/BOUND_WEIGHT)+expf(-1*d/BOUND_WEIGHT));
		float ratio=d2/d;
		for(int i=0;i<PARTITION_SPACE_DIMENSION;i++)	
			testcolorpoints[j][i]=testcolorpoints[j][i]*ratio;

	}



	float **SimilarityArray=new float*[testdatanum];
	for(long i=0;i<testdatanum;i++)
		SimilarityArray[i]= new float[traindatanum];

	for(int i=0;i<testdatanum;i++)
		for(int j=0;j<traindatanum;j++)
			SimilarityArray[i][j]=distance(testcolorpoints[i],traincolorpoints[j],PARTITION_SPACE_DIMENSION);


	//FILE *fp=fopen("predictresult.txt","w");
	int error=0;
	float TP[CLASS_NUMBER];
	float SamplesNumberofClass[CLASS_NUMBER];
	float PredictNumberofClass[CLASS_NUMBER];

	for(int i=0;i<CLASS_NUMBER;i++)
	{
		TP[i]=0;
		SamplesNumberofClass[i]=0;
		PredictNumberofClass[i]=0;
	}


	for(int i=0;i<testdatanum;i++)
	{
		int *index=new int[traindatanum];
		for(int j=0;j<traindatanum;j++)
			index[j]=j;

		for(int j=0;j<(int)(neigh_num/minweight)+2;j++)
			for(int k=traindatanum-1;k>j;k--)
			{
				if(SimilarityArray[i][index[k]]<SimilarityArray[i][index[k-1]])
				{
					int temp=index[k];
					index[k]=index[k-1];
					index[k-1]=temp;
				}
			}

		float count[CLASS_NUMBER];
		for(int j=0;j<CLASS_NUMBER;j++)
			count[j]=0;

		float sum=0;
		int j;
   		for(j=0;sum+weight[(int)(traincolorpoints[index[j]][PARTITION_SPACE_DIMENSION]+0.1F)]<neigh_num;j++)
		{
			int trainclass=(int)(traincolorpoints[index[j]][PARTITION_SPACE_DIMENSION]+0.1F);
			float w=weight[(int)(traincolorpoints[index[j]][PARTITION_SPACE_DIMENSION]+0.1F)];
			sum+=w;
			count[trainclass]+=w;
		}
		int trainclass=(int)(traincolorpoints[index[j]][PARTITION_SPACE_DIMENSION]+0.1F);
		float w=weight[(int)(traincolorpoints[index[j]][PARTITION_SPACE_DIMENSION]+0.1F)];
		count[trainclass]+=neigh_num-sum;

		int predictedclass=0;
		for(j=0;j<CLASS_NUMBER;j++)
			if(count[j]>=count[predictedclass])
				predictedclass=j;
		
		//fprintf(fp,"%d	%d\n",(int)(testcolorpoints[i][PARTITION_SPACE_DIMENSION]+0.1F),predictedclass);

		if(predictedclass!=(int)(testcolorpoints[i][PARTITION_SPACE_DIMENSION]+0.1F))
			error++;
		else
			TP[predictedclass]++;

		SamplesNumberofClass[(int)(testcolorpoints[i][PARTITION_SPACE_DIMENSION]+0.1F)]++;
		PredictNumberofClass[predictedclass]++;

		delete []index;
	}	
	//fclose(fp);

	for(long i=0;i<testdatanum;i++)
		delete []SimilarityArray[i];
	delete []SimilarityArray;
	delete []traincolorpoints;
	delete []testcolorpoints;

	measurement[0]=((float)testdatanum-error)/testdatanum;


	float AverageFMeasure=0;
	for(int i=0;i<CLASS_NUMBER;i++)
	{
		float FP=PredictNumberofClass[i]-TP[i];
		float FN=SamplesNumberofClass[i]-TP[i];

		float Precision;
		if(TP[i]==0&&FP==0)
			Precision=0;
		else 
			Precision=TP[i]/(TP[i]+FP);

		float Recall;
		if(TP[i]==0&&FN==0)
			Recall=0;
		else	
			Recall=TP[i]/(TP[i]+FN);

		float F;
		if(Precision==0&&Recall==0)
			F=0;
		else
			F=2*Precision*Recall/(Precision+Recall);
		AverageFMeasure+=F;
	}
	measurement[1]=AverageFMeasure/CLASS_NUMBER;

}


int main()
{
	srand((unsigned)time(0));
	for(int cross=0;cross<CROSS_VALIDATION_NUMBER;cross++){
	float (*train_data)[INPUT_DIMENSION+1]=new float[MAX_SAMPLES_NUMBER][INPUT_DIMENSION+1];
	float (*test_data)[INPUT_DIMENSION+1]=new float[MAX_SAMPLES_NUMBER][INPUT_DIMENSION+1];


	char trainfilename[1000];
	sprintf(trainfilename,"%s_train_%d.txt",TRAIN_FILE_NAME,cross);
	int train_sample_number=ReadFile(trainfilename, train_data);

	char testtrainfilename[1000];
	sprintf(testtrainfilename,"%s_test_%d.txt",TEST_FILE_NAME,cross);
	int test_sample_number=ReadFile(testtrainfilename, test_data);

	//printf("cross %d read file [%s %s] finished, datasize=[%d %d]!\n",cross,trainfilename,testtrainfilename,train_sample_number,test_sample_number);
		
	float weight[CLASS_NUMBER];
	GetDataWeight(train_data,train_sample_number,weight);


	int particle_dimension=NEURAL_NETWORKS_WEIGHTS_NUMBER;
	int pop_size=POPULATION_SIZE;
	
	float **pso_population_x;
	pso_population_x=new float*[pop_size];
	for(long i=0;i<pop_size;i++)
		pso_population_x[i]= new float[particle_dimension];


	float **pso_population_v;
	pso_population_v=new float*[pop_size];
	for(long i=0;i<pop_size;i++)
		pso_population_v[i]= new float[particle_dimension];

	float	*pso_population_fitness=new float[pop_size];

	float **pso_population_pbest;
	pso_population_pbest=new float*[pop_size];
	for(long i=0;i<pop_size;i++)
		pso_population_pbest[i]= new float[particle_dimension];


	float	*pso_population_pbest_fitness=new float[pop_size];
	float	*pso_population_gbest=new float[particle_dimension];
	float	pso_population_gbest_fitness=FLT_MAX;


	// Init pso population
	for(int l=0;l<pop_size;l++)
		for(int m=0;m<particle_dimension;m++)
		{
			pso_population_x[l][m]=(rand()/(float)RAND_MAX)*(INIT_MAX-INIT_MIN)+INIT_MIN;
			pso_population_v[l][m]=((rand()/(float)RAND_MAX)*(INIT_MAX-INIT_MIN)+INIT_MIN)*0.2F;
		}


	char neuralnetworkfile[1000];
	sprintf(neuralnetworkfile,"%s/neuralnetwork_%d.txt",RESULT_FILE_PATH,cross);
	FILE *fpneuralnetworkfile=fopen(neuralnetworkfile,"w");
	fclose(fpneuralnetworkfile);

	char performancefile[1000];
	sprintf(performancefile,"%s/performance_%d.txt",RESULT_FILE_PATH,cross);
	FILE *fperformancefile=fopen(performancefile,"w");
	fclose(fperformancefile);

	char finalresultsfile[1000];
	sprintf(finalresultsfile,"%s/finalresults.txt",RESULT_FILE_PATH);
	FILE *ffinalresultsfile=fopen(finalresultsfile,"w");
	fclose(ffinalresultsfile);

	// Iteration 
	for(int g=0;g<=GENERATION;g++)
	{
		#pragma omp parallel for num_threads(PARALLEL_THREAD_NUM)
		for(int l=0;l<pop_size;l++)
		{
			pso_population_fitness[l]=fitnessf(pso_population_x[l],weight,train_data,train_sample_number);
			//printf("cross%d	pso_population_fitness[%d]=%e\n",cross,l,pso_population_fitness[l]);
		}


		// Update individual best 
		for(int l=0;l<pop_size;l++)
			if(g==0||pso_population_fitness[l]<pso_population_pbest_fitness[l])
			{
				memcpy(pso_population_pbest[l],pso_population_x[l],sizeof(float)*particle_dimension);
				pso_population_pbest_fitness[l]=pso_population_fitness[l];
			}


		// Update global best
		int best_index=0;
		float best=pso_population_pbest_fitness[best_index];
		for(int l=0;l<pop_size;l++)
			if(pso_population_pbest_fitness[l]<best)
				{best_index=l;best=pso_population_pbest_fitness[l];}

		if(g==0||best<pso_population_gbest_fitness)
		{
			pso_population_gbest_fitness=best;
			memcpy(pso_population_gbest,pso_population_pbest[best_index],sizeof(float)*particle_dimension);
		}
	

		// Update velocity
		for(int l=0;l<pop_size;l++)
			for(int m=0;m<particle_dimension;m++)
			{

				pso_population_v[l][m]=((1-(float)g/GENERATION)*(0.9F-0.5F)+0.5F)*pso_population_v[l][m]
					+FI1*(rand()/(float)RAND_MAX)*(pso_population_pbest[l][m]-pso_population_x[l][m])
					+FI2*(rand()/(float)RAND_MAX)*(pso_population_gbest[m]-pso_population_x[l][m]);
				// FI0*(rand()/(float)RAND_MAX)
				if(pso_population_v[l][m]>PSO_V_MAX)
					pso_population_v[l][m]=PSO_V_MAX;
				else if(pso_population_v[l][m]<-1*PSO_V_MAX)
					pso_population_v[l][m]=-1*PSO_V_MAX;
			}

		// Update position 
		for(int l=0;l<pop_size;l++)
			for(int m=0;m<particle_dimension;m++)
				pso_population_x[l][m]=pso_population_x[l][m]+pso_population_v[l][m];

		
		if(g%10==0)
		{
			float measurement[10];
			test(pso_population_gbest,weight,train_data,train_sample_number,train_data,train_sample_number,measurement,TRAIN_PROCESS_NEIGHBOUR_NUMBER);
			//test(pso_population_gbest,weight,train_data,train_sample_number,test_data,test_sample_number,measurement,TRAIN_PROCESS_NEIGHBOUR_NUMBER);
			printf("cross %d	generation %d	fitness	%e	accuracy %e	fmeasure %e\n",cross,g,pso_population_gbest_fitness,measurement[0],measurement[1]);

			//FORMAL EXECUTION, this should be changed to
			//fpneuralnetworkfile=fopen(neuralnetworkfile,"a");
			fpneuralnetworkfile=fopen(neuralnetworkfile,"a");
			fprintf(fpneuralnetworkfile,"%d	",g);
			for(int i=0;i<particle_dimension;i++)
				fprintf(fpneuralnetworkfile,"%e	",pso_population_gbest[i]);
			fprintf(fpneuralnetworkfile,"\n");
			fclose(fpneuralnetworkfile);

			fperformancefile=fopen(performancefile,"a");
			fprintf(fperformancefile,"%d	%e	%e	%e\n",g,pso_population_gbest_fitness,measurement[0],measurement[1]);
			fclose(fperformancefile);
		}
	}



	float measurement[10];
	for(int neigh_num=MIN_NEIGHBOUR_NUMBER;neigh_num<=MAX_NEIGHBOUR_NUMBER;neigh_num+=2)
	{
		ffinalresultsfile=fopen(finalresultsfile,"a");
		test(pso_population_gbest,weight,train_data,train_sample_number,test_data,test_sample_number,measurement,neigh_num);
		fprintf(ffinalresultsfile,"%d	%d	%e	%e	%e\n",neigh_num,cross,pso_population_gbest_fitness,measurement[0],measurement[1]);
		fclose(ffinalresultsfile);
	}


	for(long i=0;i<pop_size;i++)
		delete []pso_population_x[i];
	delete []pso_population_x;

	for(long i=0;i<pop_size;i++)
		delete []pso_population_v[i];
	delete []pso_population_v;

	delete []pso_population_fitness;	


	for(long i=0;i<pop_size;i++)
		delete []pso_population_pbest[i];
	delete []pso_population_pbest;


	delete []pso_population_pbest_fitness;

	delete []pso_population_gbest;


	delete []train_data;
	delete []test_data;
	}
	return 0;
}

